import(Module_Commands)
import(Module_Control)
import(Module_DataTypes)
import(Module_Defines)
import(Module_Draw)
import(Module_Game)
import(Module_GameStates)
import(Module_Globals)
import(Module_Helpers)
import(Module_Level)
import(Module_Map)
import(Module_MapWho)
import(Module_Objects)
import(Module_Person)
import(Module_ImGui)
import(Module_Players)
import(Module_PopScript)
import(Module_System)
import(Module_Sound)
import(Module_Table)
import(Module_Math)
import(Module_Spells)
import(Module_Network)

_gsi = gsi()
_sti = scenery_type_info()
_spti = spells_type_info()
_c = constants()

function EnableSpellCharging(_pn, _spell)
  _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsNotCharging = _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsNotCharging ~ (1<<_spell-1)
end

function DisableSpellCharging(_pn, _spell)
  _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsNotCharging = _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsNotCharging | (1<<_spell-1)
end

function EnableSpell(_pn, _spell)
  _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable = _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable | (1 << _spell)
end

function DisableSpell(_pn, _spell)
  _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable = _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable ~ (1 << _spell)
end

function SetOneShotCharges(_pn, _spell, _num)
  _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsAvailableOnce[_spell] = _gsi.ThisLevelInfo.PlayerThings[_pn].SpellsAvailableOnce[_spell] | _num
end

function ResetSpellsCharging()
  for j = 0, 7 do
    for i = 0, 21 do
      if (i ~= 2 and i ~= 17) then
        DisableSpellCharging(j, i)
      end
    end
  end
end

function EnableBuilding(_pn, _building)
	 _gsi.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable = _gsi.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable | (1 << _building)
end

function DisableBuilding(_pn, _building)
	_gsi.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable = _gsi.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable ~ (1 << _building)
end

function GetTurn()
  return _gsi.Counts.GameTurn
end

function GetPlayerPeople(pn)
  return _gsi.Players[pn].NumPeople
end

function randSign()
  local a = -1
  if (G_RANDOM(2) == 0) then
    a = 1
  end
  return a
end

function everyPow(a,b)
  if (_gsi.Counts.GameTurn % a^b == 0) then
    return true else return false
  end
end

function every2Pow(a)
  if (_gsi.Counts.GameTurn % 2^a == 0) then
    return true else return false
  end
end

function DoesExist(table,input)
  for i,v in ipairs(table) do
    if (v == input) then
      return true
    end
  end

  return false
end

function GetPopLeader()
  local plr_pops = {}
  for i=0,7 do
    local pop = GetPlayerPeople(i)
    table.insert(plr_pops,pop)
  end

  local max = plr_pops[1];
  maxIndex = 1;

  for i,v in ipairs(plr_pops) do
    if (v > max) then
      maxIndex = i
      max = plr_pops[i]
    end
  end
  return maxIndex-1;
end

function tablelength(te)
  local count = 0
  for _ in pairs(te) do count = count + 1 end
  return count
end


--MAGICAL SHIELD
MagicShield = {tribe = 0}
MagicShield.__index = MagicShield

--ShieldCount is 1.5x the seconds.
function MagicShield:new(o, thing, originalMaxHealth, shieldCount)
	local o = o or {}
	setmetatable(o, MagicShield)

	o.thing = thing
	o.originalMaxHealth = originalMaxHealth
	o.timeLeftInSeconds = math.floor(shieldCount / 1.5)

	return o
end

--handle duration of shield so that max health and steel boots get readded/removed properly.
function MagicShield:handleShield()
	if (everyPow(12, 1)) then
		self.timeLeftInSeconds = self.timeLeftInSeconds - 1
	end

	if (self.thing ~= nil) then
		if (self.thing.u ~= nil) then
			if (self.thing.u.Pers ~= nil) then
				if (self.thing.u.Pers.Life <= self.originalMaxHealth or self.timeLeftInSeconds <= 0) then
					self:removeShield()
				end
			end
		end
	end

	if (self.thing ~= nil) then
		if (self.thing.u == nil) then
			DeleteMagicShieldFromList(self)
		end

		if (self.thing.u ~= nil) then
			if (self.thing.u.Pers == nil) then
				DeleteMagicShieldFromList(self)
			end
		end
	else
		DeleteMagicShieldFromList(self)
	end
end

--Reset max life to what it was before, make sure the shield flag gets removed.
function MagicShield:removeShield()
	if (self.thing ~= nil) then
		self.thing.u.Pers.MaxLife = self.originalMaxHealth

		if (self.thing.u.Pers.Life > self.originalMaxHealth) then
			self.thing.u.Pers.Life = self.originalMaxHealth
		end

		self.thing.u.Pers.u.Owned.ShieldCount = 1
		--self.thing.Flags3 = self.thing.Flags3 ~ TF3_SHIELD_ACTIVE
		self.thing.Flags2 = self.thing.Flags2 | TF2_EXTERNALLY_MOVEABLE
		self.thing.u.Pers.OnDamage:clear()
	end

	DeleteMagicShieldFromList(self)
end

function MagicShield:getShieldByPosition(D2Pos)
	if (self.thing ~= nil) then
		local c2d = Coord2D.new()
		coord3D_to_coord2D(self.thing.Pos.D3, c2d)

		if (get_world_dist_xz(D2Pos, c2d) < 32) then
			return self
		end
	end

	return nil
end

--SWAMP AND GHOSTS

Swampy = {tribe = 0}
Swampy.__index = Swampy

function Swampy:new(o, tribe, c3dLocation)
	local o = o or {}
	setmetatable(o, Swampy)

	o.tribe = tribe
	o.c3dLocation = c3dLocation
	o.initialized = 0
	o.myTiles = {}
	o.decorations = {}
	o.myLocation = Coord3D.new()

	--Have to put a reference of a reference into the actual myLocation otherwise it gets edited and swamp randomly kills itself due to shamans respawning "near it".
	--Apparently if the Xpos or other position from a Coord changes the one that has its reference ALSO changes.
	local dumbassPosX = c3dLocation.Xpos
	o.myLocation.Xpos = dumbassPosX
	local dumbassPosY = c3dLocation.Ypos
	o.myLocation.Ypos = dumbassPosY
	local dumbassPosZ = c3dLocation.Zpos
	o.myLocation.Zpos = dumbassPosZ

	o.maxTime = 2160
	o.maxKills = 10
	
	o.currentTime = 0
	o.currentKills = 0

	--o.numTilesDisabled = 0

	return o
end

function Swampy:deleteMe()
	DeleteSwampFromList(self)
	for i, tile in pairs(self.myTiles) do
		if (tile ~= nil) then
			DestroyThing(tile)
		end
	end

	for i, deco in pairs(self.decorations) do
		DestroyThing(deco)
	end
end

function Swampy:handleSwampy()
	if (self.initialized == 0) then

		for i=1, 9 do
			placeLocation = Coord3D.new()
			placeLocation = self.c3dLocation

			--Place mist
			if (i == 2) then
				placeLocation.Zpos = placeLocation.Zpos - 512
			elseif (i == 3) then
				placeLocation.Xpos = placeLocation.Xpos + 512
			elseif (i == 4) then
				placeLocation.Zpos = placeLocation.Zpos + 512
			elseif (i == 5) then
				placeLocation.Zpos = placeLocation.Zpos + 512
			elseif (i == 6) then
				placeLocation.Xpos = placeLocation.Xpos - 512
			elseif (i == 7) then
				placeLocation.Xpos = placeLocation.Xpos - 512
			elseif (i == 8) then
				placeLocation.Zpos = placeLocation.Zpos - 512
			elseif (i == 9) then
				placeLocation.Zpos = placeLocation.Zpos - 512
			end

			SearchMapCells(CIRCULAR, 0, 0, 0, world_coord3d_to_map_idx(self.c3dLocation), function(me)
          if (is_map_elem_all_sea(me) > 0) then
							
					else
						ensure_point_on_ground(placeLocation)
						local swampTile = createThing(T_EFFECT, M_EFFECT_SWAMP_MIST, self.tribe, placeLocation, false, false)
						table.insert(self.myTiles, swampTile)
						local decorationMist = createThing(T_EFFECT, M_EFFECT_SWAMP_MIST, self.tribe, placeLocation, false, false)
						table.insert(self.decorations, decorationMist)

						--Place plants
						local plantLocation = Coord3D.new()
						plantLocation = self.c3dLocation

						local locationOffset = G_RANDOM(150) - G_RANDOM(150)
						if (i == 2) then
							plantLocation.Xpos = plantLocation.Xpos + locationOffset
							local plantDecoration = createThing(T_EFFECT, M_EFFECT_REEDY_GRASS, self.tribe, placeLocation, false, false)
							table.insert(self.decorations, plantDecoration)
							plantLocation.Xpos = plantLocation.Xpos - locationOffset
						elseif (i == 4) then
							plantLocation.Xpos = plantLocation.Xpos + locationOffset
							local plantDecoration = createThing(T_EFFECT, M_EFFECT_REEDY_GRASS, self.tribe, placeLocation, false, false)
							table.insert(self.decorations, plantDecoration)
							plantLocation.Xpos = plantLocation.Xpos - locationOffset
						elseif (i == 6) then
							plantLocation.Xpos = plantLocation.Xpos + locationOffset
							local plantDecoration = createThing(T_EFFECT, M_EFFECT_REEDY_GRASS, self.tribe, placeLocation, false, false)
							table.insert(self.decorations, plantDecoration)
							plantLocation.Xpos = plantLocation.Xpos - locationOffset
						elseif (i == 8) then
							plantLocation.Xpos = plantLocation.Xpos + locationOffset
							local plantDecoration = createThing(T_EFFECT, M_EFFECT_REEDY_GRASS, self.tribe, placeLocation, false, false)
							table.insert(self.decorations, plantDecoration)
							plantLocation.Xpos = plantLocation.Xpos - locationOffset
						end
					end
					return true
					end)
		end
		queue_sound_event(self.myTiles[1], SND_EVENT_SP_SWAMP, SEF_LOOPED)

		self.initialized = 1
	end

	if (everyPow(3, 1)) then
		local unitsDamaged = 0

		for i, tile in pairs(self.myTiles) do
			if (tile ~= nil) then
				SearchMapCells(CIRCULAR, 0, 0, 0, world_coord3d_to_map_idx(tile.Pos.D3), function(me)
					me.MapWhoList:processList(function(p)
						if (p.Type == T_PERSON and is_thing_on_ground(p) == 1 and unitsDamaged < 10 and p.Owner ~= TRIBE_NEUTRAL) then
							--self.myTiles[i] = nil

							--if (p.Flags2 & TF2_THING_IS_A_GHOST_PERSON == 0) then
							--	DestroyThing(tile)
							--end
							
							--self.numTilesDisabled = self.numTilesDisabled + 1
							p.u.Pers.u.Owned.LastDamagedBy = tile.Owner
							p.u.Pers.Life = p.u.Pers.Life - 350
							unitsDamaged = unitsDamaged + 1

							if (p.u.Pers.Life <= 0 and p.Flags2 & TF2_THING_IS_A_GHOST_PERSON == 0) then
								if (p.Model == M_PERSON_MEDICINE_MAN and p.Owner ~= tile.Owner) then
									SetShamanSwampDeath(p)
								end
								self.currentKills = self.currentKills + 1
							end
						end
					return true
					end)
				return true
				end)
			end
		end

		if (self.currentKills >= self.maxKills or self.currentTime >= self.maxTime) then
			self:deleteMe()
		end
	end

	self.currentTime = self.currentTime + 1
end

Ghosts = {tribe = 0}
Ghosts.__index = Ghosts

function Ghosts:new(o, tribe, maximumGhosts)
	local o = o or {}
	setmetatable(o, Ghosts)

	o.tribe = tribe
	o.maximumGhosts = maximumGhosts

	o.ghostsTable = {}

	return o
end

function Ghosts:addGhost(thing)
	table.insert(self.ghostsTable, thing)
end

function Ghosts:killExcessGhosts()
	for i, ghst in pairs(self.ghostsTable) do
		--First clean up any dead ghosts from the table.
		if (ghst == nil or ghst.u.Pers == nil) then
			table.remove(self.ghostsTable, i)
		end
	end

	if (tableLength(self.ghostsTable) > self.maximumGhosts) then
		local amountToKill = tableLength(self.ghostsTable) - self.maximumGhosts

		for j, g in pairs(self.ghostsTable) do
			if (amountToKill > 0) then
				if (g.u.Pers ~= nil) then
					g.u.Pers.Life = 0
					table.remove(self.ghostsTable, j)
					amountToKill = amountToKill - 1
				end
			else
				break
			end
		end
	end
end

--TREES

Tree = {tribe = 0}
Tree.__index = Tree

function Tree:new(o, c2dLocation)
	local o = o or {}
	setmetatable(o, Tree)

	o.c2dLocation = c2dLocation
	--o.myPlant = myPlant
	o.myLocation = Coord2D.new()

	local dumbassPosX = c2dLocation.Xpos
	o.myLocation.Xpos = dumbassPosX
	local dumbassPosZ = c2dLocation.Zpos
	o.myLocation.Zpos = dumbassPosZ


	o.countingRegrow = 0
	o.regrowTimerSec = 0
	o.regrowOnSec = 330
	return o
end

function Tree:deleteMe()
	DeleteTreeFromList(self)
end

function Tree:startRegrowing()
	self.countingRegrow = 1
end

function Tree:showMagic()
	if (everyPow(96, 1)) then
		local c3d = Coord3D.new()
		coord2D_to_coord3D(self.myLocation, c3d)
		createThing(T_EFFECT, M_EFFECT_SPARKLE, TRIBE_NEUTRAL, c3d, false, false)
	end
end

--Call me every second (12 turns) for performance
function Tree:handleTree()
	if (self.countingRegrow == 1) then
		self.regrowTimerSec = self.regrowTimerSec + 1

		local c3d = Coord3D.new()
		coord2D_to_coord3D(self.myLocation, c3d)

		SearchMapCells(CIRCULAR, 0, 0, 0, world_coord3d_to_map_idx(c3d), function(me)
        if (me.Flags & (1<<9) >= 1) then
						self:deleteMe()
						return false
        end
    return true
    end)

		if (self.regrowTimerSec >= self.regrowOnSec) then
			local c3d = Coord3D.new()
			coord2D_to_coord3D(self.myLocation, c3d)
			
			local treeModel = G_RANDOM(6)+1
			local rotation = G_RANDOM(2048)+1

			if (treeModel == 1) then
				local tree = createThing(T_SCENERY, M_SCENERY_TREE_1, TRIBE_NEUTRAL, c3d, false, false)
				tree.u.Scenery.ResourceRemaining = 100
				tree.AngleXZ = rotation
			end
			if (treeModel == 2) then
				local tree = createThing(T_SCENERY, M_SCENERY_TREE_2, TRIBE_NEUTRAL, c3d, false, false)
				tree.u.Scenery.ResourceRemaining = 100
				tree.AngleXZ = rotation
			end
			if (treeModel == 3) then
				local tree = createThing(T_SCENERY, M_SCENERY_TREE_3, TRIBE_NEUTRAL, c3d, false, false)
				tree.u.Scenery.ResourceRemaining = 100
				tree.AngleXZ = rotation
			end
			if (treeModel == 4) then
				local tree = createThing(T_SCENERY, M_SCENERY_TREE_4, TRIBE_NEUTRAL, c3d, false, false)
				tree.u.Scenery.ResourceRemaining = 100
				tree.AngleXZ = rotation
			end
			if (treeModel == 5) then
				local tree = createThing(T_SCENERY, M_SCENERY_TREE_5, TRIBE_NEUTRAL, c3d, false, false)
				tree.u.Scenery.ResourceRemaining = 100
				tree.AngleXZ = rotation
			end
			if (treeModel == 6) then
				local tree = createThing(T_SCENERY, M_SCENERY_TREE_6, TRIBE_NEUTRAL, c3d, false, false)
				tree.u.Scenery.ResourceRemaining = 100
				tree.AngleXZ = rotation
			end

			self.countingRegrow = 0
			self.regrowTimerSec = 0
		end
	end
end

--VEHICLES

Vehicle = {tribe = 0}
Vehicle.__index = Vehicle

function Vehicle:new(o, tribe, maxHealth, vehicleThing)
	local o = o or {}
	setmetatable(o, Vehicle)

	o.tribe = tribe
	o.maxHealth = maxHealth
	o.currentHealth = maxHealth
	o.vehicleThing = vehicleThing

	o.fireballsGround = {}
	o.fireballsTower = {}

	o.shamanIsInsideMe = 0
	o.lastDamagedBy = -1
	--o.lastNumOccupants = 0

	return o
end

function Vehicle:handleVehicle()
	if (self.currentHealth <= 0) then
		if (self.vehicleThing ~= nil) then
			self.vehicleThing.u.Vehicle.Life = -1
			self.vehicleThing = nil
			DeleteVehicleFromList(self)
		end
	end

	if (self.shamanIsInsideMe == 1 and self.vehicleThing ~= nil) then
		if (self.vehicleThing.Move.SelfPowerSpeed >= 45) then
			self.vehicleThing.Move.SelfPowerSpeed = 45
		end
	end

	for i, ball in pairs(self.fireballsGround) do
		if (GetTurn() >= ball) then
			self:damageVehicle(50)
			table.remove(self.fireballsGround, i)
		end
	end

	for i, ball in pairs(self.fireballsTower) do
		if (GetTurn() + 6 >= ball) then
			if (self.shamanIsInsideMe == 1) then
				if (self.vehicleThing ~= nil) then
					local targetSham = getShaman(self.vehicleThing.Owner)

					if (targetSham ~= nil) then
						if (targetSham.u ~= nil) then
							if (targetSham.u.Pers ~= nil) then
								targetSham.u.Pers.Life = 800
							end
						end
					end
				end
			end
		end
		
		if (GetTurn() >= ball) then
			self:damageVehicle(175)
			table.remove(self.fireballsTower, i)
		end
	end

	if (self.vehicleThing ~= nil) then
		myOwnersShaman = getShaman(self.vehicleThing.Owner)

		if (myOwnersShaman ~= nil and self.shamanIsInsideMe == 1) then
			if (myOwnersShaman.Flags2 & TF2_IN_AIRSHIP <= 0 and is_person_in_boat(myOwnersShaman) == 0) then
				self:SetShamanInsideMe(0)
			end
		end
	end
	--Check if a shaman is inside me but only do it when the number of occupants has changed since last checking.
	--[[if (self.vehicleThing ~= nil) then
		if (self.vehicleThing.u ~= nil) then
			if (self.vehicleThing.u.Vehicle ~= nil) then
				if (self.vehicleThing.u.Vehicle.NumOccupants < self.lastNumOccupants or self.vehicleThing.u.Vehicle.NumOccupants > self.lastNumOccupants) then
					self.lastNumOccupants = self.vehicleThing.u.Vehicle.NumOccupants
					local foundShaman = 0
					log_msg(TRIBE_NEUTRAL, "NumOccupants Changed")
					SearchMapCells(CIRCULAR, 0, 0, 0, world_coord3d_to_map_idx(self.vehicleThing.Pos.D3), function(me)
						me.MapWhoList:processList(function(p)
							if (p.Type == T_PERSON) then
								log_msg(TRIBE_NEUTRAL, "Person "..p.Model)
								if (p.Model == M_PERSON_MEDICINE_MAN) then  
									self.shamanIsInsideMe = 1    
									log_msg(TRIBE_NEUTRAL, "Shaman Inside")
									foundShaman = 1
								end
							end
						return true
						end)
					return true
					end)

					if (foundShaman == 0) then
						self.shamanIsInsideMe = 0
						log_msg(TRIBE_NEUTRAL, "Shaman Outside")
					end
				end
			end
		end
	end--]]

	if (everyPow(12, 1)) then
		if (self.currentHealth <= self.maxHealth) then
			self.currentHealth = self.currentHealth + 32
		end
	end
end

function Vehicle:SetShamanInsideMe(shamanInside)
	self.shamanIsInsideMe = shamanInside
end

function Vehicle:AddFireball(turn, tower, playerNrSource)
	self.lastDamagedBy = playerNrSource

	if (tower == 1) then
		table.insert(self.fireballsTower, turn)
	else
		table.insert(self.fireballsGround, turn)
	end
end

function Vehicle:damageVehicle(dmg)
	self.currentHealth = self.currentHealth - dmg
end

function Vehicle:getVehicleByPosition(D2Pos)
	if (self.vehicleThing ~= nil) then
		local c2d = Coord2D.new()
		coord3D_to_coord2D(self.vehicleThing.Pos.D3, c2d)

		if (get_world_dist_xz(D2Pos, c2d) < 160) then
			return self
		end
	end

	return nil
end